# -*- coding: utf-8 -*-

# from .helcats import ICMECAT_EVENT  # noqa: F401
from .satellites import BEPI, DSCOVR, MES, PSP, SOLO, STA, STB, VEX, WIND  # noqa: F401
from .spice import SpiceObject as _SpiceObject
from .util import get_heliosat_paths


__author__ = "Andreas J. Weiss"
__copyright__ = "Copyright (C) 2019 Andreas J. Weiss"
__license__ = "MIT"
__version__ = "0.4.10"

_paths = get_heliosat_paths()
_spice = None

# common solar system objects
Sun = lambda: _SpiceObject(None, "SUN", skip_init=True)  # noqa: E731
Mercury = lambda: _SpiceObject(None, "MERCURY", skip_init=True)  # noqa: E731
Venus = lambda: _SpiceObject(None, "VENUS", skip_init=True)  # noqa: E731
Earth = lambda: _SpiceObject(None, "EARTH", skip_init=True)  # noqa: E731
Moon = lambda: _SpiceObject(None, "MOON", skip_init=True)  # noqa: E731
Mars = lambda: _SpiceObject(None, "MARS BARYCENTER", skip_init=True)  # noqa: E731
Jupiter = lambda: _SpiceObject(None, "JUPITER BARYCENTER", skip_init=True)  # noqa: E731
Saturn = lambda: _SpiceObject(None, "SATURN BARYCENTER", skip_init=True)  # noqa: E731
Uranus = lambda: _SpiceObject(None, "URANUS BARYCENTER", skip_init=True)  # noqa: E731
Neptune = lambda: _SpiceObject(None, "NEPTUNE BARYCENTER", skip_init=True)  # noqa: E731



# -*- coding: utf-8 -*-

"""spice.py

Implements the SPICE base class and SPICE related functions.
"""

import datetime
import heliosat
import json
import logging
import numpy as np
import os
import requests
import spiceypy
import time

from heliosat.download import download_files, urls_expand
from typing import Iterable, List, Optional, Union


class SpiceObject(object):
    """Base SPICE class for SPICE aware objects (bodies, spacecraft etc.).
    """
    def __init__(self, name: str, body_name: str, kernel_group: Optional[str] = None,
                 skip_init: bool = False, force_download: bool = False):
        """Initialize SPICE aware object and load required kernels if given.

        Parameters
        ----------
        name : str
            Object name (for spacecrafts as defined in the spacecraft.json file).
        body_name : str
            SPICE object name.
        kernel_group : str, optional
            SPICE kernel group name, by default None.
        skip_init: bool, optional
            Initialize default spice kernels, by default True.
        force_download: bool, optional
            Force kernel downloads (otherwise skip remote checks), by default False.
        """
        logger = logging.getLogger(__name__)

        if skip_init is False and heliosat._spice is None:
            logger.info("running spice_init")
            spice_init(force_download)

        self.name = name
        self.body_name = body_name

        if kernel_group:
            spice_load(kernel_group, force_download)

    def trajectory(self, t: Union[datetime.datetime, Iterable[datetime.datetime]],
                   frame: str = "J2000", observer: str = "SUN", units: str = "AU"):
        """Evaluate body trajectory at given datetimes.

        Parameters
        ----------
        t : Union[datetime.datetime, Iterable[datetime.datetime]]
            Evaluate datetime(s).
        frame : str, optional
            Trajectory reference frame, by default "J2000".
        observer : str, optional
            Observer body name, by default "SUN".
        units : str, optional
            Output units, by default "AU".

        Returns
        -------
        np.ndarray
            Body trajectory.

        Raises
        ------
        NotImplementedError
            If units are invalid.
        """
        logger = logging.getLogger(__name__)

        if heliosat._spice is None:
            logger.info("running spice_init")
            spice_init(False)

        trajectory = np.array(
            spiceypy.spkpos(
                self.body_name,
                spiceypy.datetime2et(t),
                frame,
                "NONE",
                observer
            )[0]
        )

        if units == "AU":
            trajectory *= 6.68459e-9
        elif units == "m":
            trajectory *= 1e3
        elif units == "km":
            pass
        else:
            logger.exception("unit \"%s\" is not supported", units)
            raise NotImplementedError("unit \"%s\" is not supported", units)

        return trajectory


def spice_init(force_download: bool = False):
    """Initialize SPICE kernels.

    Parameters
    ----------
    force_download: bool, optional
            Force kernel downloads (otherwise skip remote checks), by default False.
    """
    logger = logging.getLogger(__name__)

    # clear all
    spiceypy.kclear()

    kernels_path = heliosat._paths["kernels"]

    json_kernels_path = os.path.join(os.path.dirname(heliosat.__file__), "json/kernels.json")

    heliosat._spice = {"kernel_groups": None, "kernels_loaded": None}

    # json/kernels.json
    logger.info("loading available kernels")

    with open(json_kernels_path) as json_file:
        json_kernels = json.load(json_file)

    timestamp = json_kernels.get("timestamp", 0)

    # update kernels if timestamp is older than a day
    if os.path.exists(kernels_path) and timestamp > time.time() - 86400:
        logger.info("skipping check for new kernels")
    else:
        if not os.path.exists(kernels_path):
            os.makedirs(kernels_path)

        try:
            logger.info("checking for new kernels")

            json_kernels["_groups"] = dict(json_kernels["groups"])

            for group in json_kernels["_groups"]:
                json_kernels["_groups"][group] = \
                    urls_expand(json_kernels["_groups"][group])

            json_kernels["timestamp"] = time.time()

            with open(json_kernels_path, "w") as json_file:
                json.dump(json_kernels, json_file, indent=4)
        except requests.RequestException as requests_error:
            logger.error("skipping check for new kernels (%s)", requests_error)

            if "_groups" not in json_kernels:
                logger.exception("expanded kernel group \"_groups\" could not be generated"
                                 "and previous version cannot be found")
                raise RuntimeError("expanded kernel group \"_groups\" could not be generated"
                                   "and previous version cannot be found")

    # load generic kernels
    download_files(json_kernels["generic"], kernels_path, force=force_download, logger=logger)

    heliosat._spice["kernel_groups"] = json_kernels["_groups"]
    heliosat._spice["kernels_loaded"] = []

    for kernel_url in json_kernels["generic"]:
        spiceypy.furnsh(os.path.join(kernels_path, kernel_url.split("/")[-1]))
        heliosat._spice["kernels_loaded"].append(kernel_url)


def spice_load(kernel_group: str, force_download: bool = False):
    """Load SPICE kernel group.

    Parameters
    ----------
    kernel_group : str
        SPICE kernel group name as defined in the spacecraft.json file.
    force_download: bool, optional
            Force kernel downloads (otherwise skip remote checks), by default False.
    """
    logger = logging.getLogger(__name__)

    kernels_path = heliosat._paths["kernels"]

    if kernel_group:
        kernels_required = heliosat._spice["kernel_groups"][kernel_group]

        for kernel in list(kernels_required):
            if kernel in heliosat._spice["kernels_loaded"]:
                kernels_required.remove(kernel)

        if len(kernels_required) == 0:
            return

        logger.debug("loading kernel group \"%s\"", kernel_group)

        download_files(kernels_required, kernels_path, force=force_download, logger=logger)

        for kernel_url in kernels_required:
            kernel = os.path.join(kernels_path, kernel_url.split("/")[-1])

            logger.debug("loading kernel \"%s\"", kernel)

            try:
                spiceypy.furnsh(kernel)
                heliosat._spice["kernels_loaded"].append(kernel_url)
            except spiceypy.support_types.SpiceyError:
                logger.exception("failed to load kernel \"%s\"", kernel_url)


def spice_reload(kernel_urls: List[str], force_download: bool = False):
    """Load SPICE kernels by url. This function is meant to be used in child processes that need
    to reload all SPICE kernels.

    Parameters
    ----------
    kernel_urls : List[str]
        SPICE kernel urls.
    force_download: bool, optional
            Force kernel downloads (otherwise skip remote checks), by default False.
    """
    logger = logging.getLogger(__name__)

    kernels_path = heliosat._paths["kernels"]

    kernels_required = kernel_urls
    
    for kernel in list(kernels_required):
        if kernel in heliosat._spice["kernels_loaded"]:
            kernels_required.remove(kernel)

    if len(kernels_required) == 0:
        return

    download_files(kernels_required, kernels_path, force=force_download, logger=logger)

    for kernel_url in kernels_required:
        kernel = os.path.join(kernels_path, kernel_url.split("/")[-1])

        logger.debug("loading kernel \"%s\"", kernel)
        try:
            spiceypy.furnsh(kernel)
            heliosat._spice["kernels_loaded"].append(kernel_url)
        except spiceypy.support_types.SpiceyError:
            logger.exception("failed to load kernel \"%s\"", kernel_url)



# -*- coding: utf-8 -*-

"""download.py

Utility functions for downloading data files. Intended for internal use only.
"""

import logging
import numpy as np
import os
import re
import requests
import requests_ftp


from bs4 import BeautifulSoup
from concurrent.futures import ThreadPoolExecutor
from typing import List, Union


def download_files(file_urls: List[str], file_paths: Union[str, List[str]], **kwargs) -> List[bool]:
    """Download files from given url's and store them locally.

    Parameters
    ----------
    file_urls : List[str]
        Target url's.
    file_paths : Union[str, list[str]]
        Destination file paths (or optionally the destination folder).

    Other Parameters
    ----------------
    force: bool
        Force overwrite, by default False.
    logger : logging.Logger
        Logger handle, by default None.
    threads: int
        Number of parallel threads, by default 32.

    Returns
    -------
    List[bool]
        Flags if files were downloaded succesfully, or they already existed.

    Raises
    ------
    ValueError
        If file_paths is not a folder and the number of given file_paths does not match the number
        of given url's.
    """
    force = kwargs.get("force", kwargs.get("overwrite", False))
    logger = kwargs.get("logger", logging.getLogger(__name__))
    threads = min(len(file_urls), kwargs.get("threads", 32))

    if isinstance(file_paths, str):
        if not os.path.exists(file_paths):
            os.makedirs(file_paths)

        file_paths = [os.path.join(file_paths, file_url.split("/")[-1]) for file_url in file_urls]

    if len(file_urls) != len(file_paths):
        logger.exception("invalid file path list (length mismatch)")
        raise ValueError("invalid file path list (length mismatch)")

    with ThreadPoolExecutor(max_workers=threads) as executor:
        futures = executor.map(_worker_download_files, [(file_urls[i], file_paths[i], force, logger)
                                                        for i in range(len(file_urls))])

    results = [_ for _ in futures]

    if np.any(np.invert(results)):
        logger.warning("checked/downloaded %i/%i files", np.sum(results), len(results))
    else:
        logger.info("checked/downloaded %i files", len(results))

    return results


def _worker_download_files(args: tuple) -> bool:
    """Worker function for downloading files.

    Parameters
    ----------
    args: (str, str, bool, logging.Logger)
        Function arguments as tuple.

    Returns
    -------
    bool
        Flag if file was downloaded, or already existed.

    Raises
    ------
    NotImplementedError
        If url is not http(s) or ftp.
    """
    try:
        (file_url, file_path, force, logger) = args

        file_already_exists = os.path.isfile(file_path)

        if file_already_exists and os.path.getsize(file_path) == 0:
            file_already_exists = False

        if not force and file_already_exists:
            # skip download if file already exists
            logger.info("skipped \"%s\"", file_url)
            return True

        if file_url.startswith("http"):
            logger.debug("downloading \"%s\"", file_url)
            response = requests.get(file_url)
        elif file_url.startswith("ftp"):
            logger.debug("downloading (ftp) \"%s\"", file_url)
            requests_ftp.monkeypatch_session()
            s = requests.Session()
            response = s.get(file_url)
            s.close()
        else:
            logger.exception("invalid url \"%s\"", file_url)
            raise NotImplementedError("invalid url \"%s\"", file_url)

        if response.ok:
            # fix for url's that return 200 instead of a 404
            if "Content-Length" in response.headers and \
                    int(response.headers.get("Content-Length")) < 1000:
                raise requests.HTTPError("Content-Length is very small"
                                         "(url is most likely not a valid file)")
            else:
                with open(file_path, "wb") as file:
                    file.write(response.content)
        else:
            return response.raise_for_status()

        return True
    except requests.RequestException as error:
        if file_already_exists:
            logger.error("failed to check existing file \"%s\" (%s)", file_url, error)

            # local file is assumed to be correct if it cannot be checked
            return True
        else:
            logger.error("failed to download \"%s\" (%s)", file_url, error)

            # delete empty file that may have been created
            if os.path.exists(file_path):
                os.remove(file_path)

        return False


def urls_expand(urls: List[str], pre: str = "$") -> list:
    """"Expand list of url's using regular expressions. Both HTTP and FTP url's are supported.

    Parameters
    ----------
    urls : List[str]
        Url list (including expandables and non-expandables).
    pre : str
        Prefix to identify expandable url's.

    Returns
    -------
    List[str]
        Expanded url list (including non-expandables).

    Raises
    ------
    NotImplementedError
        If url is not http(s) or ftp.
    """
    logger = logging.getLogger(__name__)

    # copy url's that do not need expansion
    urls_expanded = [_ for _ in urls if not _.startswith("$")]

    for url in [_ for _ in urls if _.startswith(pre)]:
        if url[1:].startswith("http"):
            url_parent = "/".join(url[1:].split("/")[:-1])
            url_regex = url.split("/")[-1]

            response = requests.get(url_parent)

            if response.ok:
                response_text = response.text
            else:
                return response.raise_for_status()

            # match all urls's with regex pattern
            soup = BeautifulSoup(response_text, "html.parser")

            for url_child in [_.get("href") for _ in soup.find_all("a")]:
                if url_child and re.match(url_regex, url_child):
                    urls_expanded.append("/".join([url_parent, url_child]))
        elif url[1:].startswith("ftp"):
            raise NotImplementedError("ftp is currently not support for expansion")
        else:
            logger.exception("failed to expand url: \"%s\"", url)
            raise NotImplementedError("failed to expand url: \"%s\"", url)

    return urls_expanded


def urls_resolve(urls: List[str]) -> List[str]:
    """Resolve list of url's using regular expressions.

    Parameters
    ----------
    urls : List[str]
        Url list.

    Returns
    -------
    List[str]
        Resolved url list.
    """
    # organize url's so that any page is only called once
    url_parents = {"/".join(url.split("/")[:-1]): [] for url in urls}

    for url in urls:
        url_parents["/".join(url.split("/")[:-1])].append(url.split("/")[-1])

    urls_resolved = []

    for url_parent in url_parents:
        response = requests.get(url_parent)

        if response.ok:
            response_text = response.text
        else:
            return response.raise_for_status()

        # match all url's with regex pattern
        soup = BeautifulSoup(response_text, "html.parser")
        hrefs = [_.get("href") for _ in soup.find_all("a")]

        for url_regex in url_parents[url_parent]:
            # incase of multiple matches, choose the last one
            last_match = None

            for url_child in hrefs:
                if url_child and re.match(url_regex, url_child):
                    last_match = "/".join([url_parent, url_child])

            if last_match:
                urls_resolved.append(last_match)

    return urls_resolved



# -*- coding: utf-8 -*-

"""spacecraft.py

Implements the Spacecraft base class. All implemented spacecraft classes inherit from the base
Spacecraft class which defines basic data handling routines.
"""

import cdflib
import datetime
import gzip
import heliosat
import json
import logging
import multiprocessing
import numpy as np
import os
import shutil

from astropy.io import fits
from concurrent.futures import ProcessPoolExecutor
from heliosat.caching import cache_add_entry, cache_entry_exists, \
    cache_generate_key, cache_get_entry
from heliosat.coordinates import transform_pos
from heliosat.download import download_files, urls_resolve
from heliosat.smoothing import smooth_data
from heliosat.spice import SpiceObject, spice_load
from heliosat.util import datetime_utc, datetime_utc_timestamp, datetime_to_string, \
    sanitize_datetimes, string_to_datetime
from netCDF4 import Dataset
from itertools import compress


class Spacecraft(SpiceObject):
    """Base Spacecraft class.
    """
    def __init__(self, name, body_name, **kwargs):
        """Initialize Spacecraft object. This just fetches the spacecraft.json entry and loads any
        spacecraft specific SPICE kernels.

        Parameters
        ----------
        name : str
            Spacecraft name as defined in the spacecraft.json file.
        body_name : str
            SPICE body name assosicated with the given spacecraft. In some cases a different body
            is given with a similar orbital trajectory (e.g. Earth for DSCOVR/Wind).

        Other Parameters
        ----------------
        kernel_group: str
            Spacecraft kernel group as defined in the spacecraft.json file, by default None.
        force_download: bool, optional
            Force kernel downloads (otherwise skip remote checks), by default False.

        Raises
        ------
        NotImplementedError
            If spacecraft is not implemented (not defined in the spacecraft.json file).
        """
        logger = logging.getLogger(__name__)

        super(Spacecraft, self).__init__(name, body_name, **kwargs)

        if heliosat._spice.get("spacecraft", None) is None:
            logger.info("loading available spacecraft")

            module_path = os.path.dirname(heliosat.__file__)

            with open(os.path.join(module_path, "json/spacecraft.json")) as json_file:
                heliosat._spice["spacecraft"] = json.load(json_file)

            self.spacecraft = heliosat._spice["spacecraft"][name]
        elif name in heliosat._spice["spacecraft"]:
            self.spacecraft = heliosat._spice["spacecraft"][name]
        else:
            logger.exception("spacecraft \"%s\" is not implemented", name)
            raise NotImplementedError("spacecraft \"%s\" is not implemented", name)

        if kwargs.get("kernel_group", None) is None and self.spacecraft.get("kernel_group", None):
            spice_load(self.spacecraft["kernel_group"],
                       force_download=kwargs.get("force_download", False))

    def get_data(self, t, data_key, **kwargs):
        """Get processed data for type "data_key" at specified datetimes t.

        Parameters
        ----------
        t : list[datetime.datetime]
            Evaluation datetimes.
        data_key : str
            Data key.

        Other Parameters
        ----------------
        cache: bool
            Cache data, by default False.
        cache_raw: bool
            Store raw data files, by default False.
        columns: list[str]
            Read columns instead of default, by default None.
        extra_columns: list[str]
            Read extra columns ontop of default, by default None.
        frame: str
            Data reference frame, by default None.
        frame_cadence: float
            Evaluate frame transformation matrix every "frame_cadence" seconds instead of at very
            time point (significant speed up).+
        remove_nans : bool
            Remove invalid values, by default False.
        return_datetimes: bool
            Return datetimes instead of timestamps, by default False.
        sampling_freq : int
            Sampling frequency in seconds if t is only given by start/end, by default 60.
        smoothing: str
            Smoothing method, by default None.
        smoothing_scale: float
            Smoothing scale in seconds, by default 300.
        force_download: bool, optional
            Force kernel downloads (otherwise skip remote checks), by default False.

        Returns
        -------
        (list[float], np.ndarray)
            Evaluation datetimes as timestamps & processed data array.
        """
        logger = logging.getLogger(__name__)

        data_key = self.resolve_data_key(data_key)

        # sanitize time input
        t = sanitize_datetimes(t)

        if isinstance(t, datetime.datetime):
            identifiers = {
                "data_key": data_key,
                "spacecraft": self.name,
                "times": t.timestamp(),
                "version": heliosat.__version__,
                "remove_nans": kwargs.get("remove_nans", False),
                "sampling_freq": kwargs.get("sampling_freq", 60)
            }
        else:
            identifiers = {
                "data_key": data_key,
                "spacecraft": self.name,
                "times": [_t.timestamp() for _t in t],
                "version": heliosat.__version__,
                "remove_nans": kwargs.get("remove_nans", False),
                "sampling_freq": kwargs.get("sampling_freq", 60)
            }

        identifiers.update(kwargs)

        # clean up kwargs so it can be passed on to get_data_raw
        cache = kwargs.get("cache", False)
        remove_nans = kwargs.pop("remove_nans", False)
        return_datetimes = kwargs.pop("return_datetimes", False)
        sampling_freq = kwargs.pop("sampling_freq", 60)
        smoothing = kwargs.get("smoothing", None)
        force_download = kwargs.get("force_download", False)

        if cache:
            # fetch cached entry
            cache_key = cache_generate_key(identifiers)

            if cache_entry_exists(cache_key):
                logger.info("loading cache entry \"%s\"", cache_key)
                return cache_get_entry(cache_key)
            else:
                logger.info("cache entry \"%s\" not found", cache_key)

        # if time is 2 elements, build list
        if len(t) == 2:
            time_ts = np.linspace(t[0].timestamp(), t[1].timestamp(),
                                  int((t[1].timestamp() - t[0].timestamp()) // sampling_freq))

            t = [datetime.datetime.fromtimestamp(_, datetime.timezone.utc) for _ in time_ts]

        time, data = self.get_data_raw(t[0], t[-1], data_key, **kwargs)

        if smoothing:
            # smooth data
            # all parameters that include "smoothing" are passed onto the smoothing function
            smoothing_dict = {"smoothing": smoothing}

            logger.info("smoothing data (%s)", smoothing)

            for key in dict(kwargs):
                if "smoothing" in key:
                    smoothing_dict[key] = kwargs.pop(key)

            time, data = smooth_data(t, time, data, **smoothing_dict)
        else:
            time, data = smooth_data(t, time, data, smoothing="closest")

        if return_datetimes:
            _time = list(time)

            for i in range(len(_time)):
                if _time[i] != np.nan:
                    _time[i] = datetime_utc_timestamp(time[i])

            time = np.array(_time)

        if remove_nans:
            nanfilter = np.invert(np.isnan(data[:, 0]))

            time = time[nanfilter]
            data = data[nanfilter]

        if cache and not cache_entry_exists(cache_key):
            # save cache entry
            logger.info("generating cache entry \"%s\"", cache_key)
            cache_add_entry(cache_key, (time, data))

        return time, data

    def get_data_raw(self, range_start, range_end, data_key, **kwargs):
        """Get raw data for type "data_key" within specified time range.

        Parameters
        ----------
        range_start : datetime.datetime
            Time range start datetime.
        range_end : datetime.datetime
            Time range end datetime.
        data_key : str
            Data key.

        Other Parameters
        ----------------
        cache_raw: bool
            Store raw data files, by default True.
        columns: list[str]
            Read columns instead of default, by default None.
        extra_columns: list[str]
            Read extra columns ontop of default, by default None.
        frame: str
            Data reference frame, by default None.
        frame_cadence: float
            Evaluate frame transformation matrix every "frame_cadence" seconds instead of at very
            time point (significant speed up).
        return_datetimes: bool
            Return datetimes instead of timestamps, by default False.
        force_download: bool, optional
            Force kernel downloads (otherwise skip remote checks), by default False.

        Returns
        -------
        (np.ndarray, np.ndarray)
            Evaluation datetimes as timestamps & raw data array.

        Raises
        ------
        KeyError
            If data_key does not exist.
        """
        logger = logging.getLogger(__name__)

        data_key = self.resolve_data_key(data_key)

        range_start = sanitize_datetimes(range_start)
        range_end = sanitize_datetimes(range_end)

        frame = kwargs.get("frame", None)
        frame_cadence = kwargs.get("frame_cadence", None)
        force_download = kwargs.get("force_download", False)

        files, file_versions = self.get_data_files(range_start, range_end, data_key,
                                                   force_download=force_download)

        logger.info("using %s files to generate "
                    "data in between %s - %s", len(files), range_start, range_end)

        max_workers = min(multiprocessing.cpu_count() * 2, len(files))

        # setup columns, ~ is a placeholder for the default columns
        columns = kwargs.get("columns", ["~"])
        columns.extend(kwargs.get("extra_columns", []))

        kernels = heliosat._spice["kernels_loaded"]

        with ProcessPoolExecutor(max_workers=max_workers) as executor:
            args = [(files[i], file_versions[i], range_start, range_end, columns, frame,
                     frame_cadence, kernels)
                    for i in range(len(files))]

            futures = executor.map(read_task, args)

        results = [_ for _ in futures if _]

        time_all = np.concatenate([_[0] for _ in results])
        data_all = np.concatenate([_[1] for _ in results])

        # delete raw data files if required (must be explicitly set)
        if not kwargs.get("cache_raw", True):
            for file in files:
                os.remove(file)

        if kwargs.get("return_datetimes", False):
            _time = list(time_all)

            for i in range(len(_time)):
                if _time[i] != np.nan:
                    _time[i] = datetime_utc_timestamp(_time[i])

            time_all = _time

        return time_all, data_all

    def get_data_files(self, range_start, range_end, data_key, force_download=False):
        """Get raw data file paths for type "data_key" within specified time range.

        Parameters
        ----------
        range_start : datetime.datetime
            Time range start datetime.
        range_end : datetime.datetime
            Time range end datetime.
        data_key : str
            Data key.
        force_download: bool, optional
            Force kernel downloads (otherwise skip remote checks), by default False.

        Returns
        -------
        list
            Raw data file paths.
        """
        logger = logging.getLogger(__name__)

        if range_end < range_start:
            raise ValueError("starting date must be after final date")

        data_key = self.resolve_data_key(data_key)
        data_path = os.path.join(heliosat._paths["data"], data_key)

        urls = []
        versions = []

        # adjust ranges slightly
        range_start -= datetime.timedelta(hours=range_start.hour, minutes=range_start.minute,
                                          seconds=range_start.second)
        if range_end.hour == 0 and range_end.minute == 0 and range_end.second == 0:
            range_end -= datetime.timedelta(seconds=1)

        # build url for each day in range (this assumes all files are daily)
        for day in [range_start + datetime.timedelta(days=i)
                    for i in range((range_end - range_start).days + 1)]:
            url = self.spacecraft["data_keys"][data_key]["urls"]

            url = url.replace("{YYYY}", str(day.year))
            url = url.replace("{YY}", "{0:02d}".format(day.year % 100))
            url = url.replace("{MM}", "{:02d}".format(day.month))
            url = url.replace("{MONTH}", day.strftime("%B")[:3].upper())
            url = url.replace("{DD}", "{:02d}".format(day.day))
            url = url.replace("{DOY}", "{:03d}".format(day.timetuple().tm_yday))

            doym1 = datetime_utc(day.year, day.month, 1)

            if day.month == 12:
                doym2 = datetime_utc(day.year + 1, 1, 1) - datetime.timedelta(days=1)
            else:
                doym2 = datetime_utc(day.year, day.month + 1, 1) - datetime.timedelta(days=1)

            url = url.replace("{DOYM1}", "{:03d}".format(doym1.timetuple().tm_yday))
            url = url.replace("{DOYM2}", "{:03d}".format(doym2.timetuple().tm_yday))

            # insert version if required
            version = self.get_data_version(data_key, day)

            for i in range(len(version.get("identifiers", []))):
                url = url.replace("{{V{0}}}".format(i), version["identifiers"][i])

            urls.append(url)
            versions.append(version)

        # resolve regex urls if required
        if self.spacecraft["data_keys"][data_key].get("use_regex", False):
            urls = urls_resolve(urls)

        # some files are archives, skip the download if the extracted versions exist
        files_extracted = []
        versions_extracted = []

        for url in list(urls):
            if url.endswith(".gz"):
                file = ".".join(url.split("/")[-1].split(".")[:-1])

                if os.path.exists(os.path.join(data_path, file)):
                    logger.info("skipping download for \"%s\" as extracted version already exists",
                                url)

                    # remove url & version, and append extracted file at end
                    rindex = urls.index(url)

                    files_extracted.append(os.path.join(data_path, file))
                    versions_extracted.append(versions[rindex])

                    del urls[rindex]
                    del versions[rindex]

        if len(urls) > 0:
            results = download_files(urls, data_path, force=force_download, logger=logger)
            check_downloads = True
        else:
            check_downloads = False

        files = [
            os.path.join(data_path, urls[i].split("/")[-1])
            for i in range(0, len(urls))
        ]

        # extract new archives and add old ones to the list
        files.extend(files_extracted)
        versions.extend(versions_extracted)

        for file in list(files):
            if file.endswith(".gz"):
                with gzip.open(file, "rb") as file_gz:
                    with open(".".join(file.split(".")[:-1]), "wb") as file_extracted:
                        shutil.copyfileobj(file_gz, file_extracted)

                os.remove(file)
                rindex = files.index(file)
                files[rindex] = ".".join(file.split(".")[:-1])

        if check_downloads:
            return list(compress(files, results)), list(compress(versions, results))
        else:
            return files, versions

    def get_data_version(self, data_key, time=None):
        """Get version information from type "data_key" at specific time.

        Parameters
        ----------
        data_key : str
            Data key.
        time : datetime.datetime
            Version time, by default None.

        Returns
        -------
        dict
            Version information.

        Raises
        ------
        RuntimeError
            If no version is defined at specified time.
        """
        data_key = self.resolve_data_key(data_key)

        versions = self.spacecraft["data_keys"][data_key]["versions"]

        ver_found = False

        # get version defaults
        selected_version = dict(self.spacecraft["data_keys"][data_key]["version_default"])

        if time is None:
            return selected_version

        for version in versions:
            if string_to_datetime(version["version_start"]) <= time < \
               string_to_datetime(version["version_end"]):
                selected_version.update(version)

                return selected_version

        if not ver_found:
            raise RuntimeError("no version found for data key \"%s\" at time %s", data_key,
                               datetime_to_string(time))

    def get_data_columns(self, data_key, time=None, show_all=False):
        """Get data columns from type "data_key" at specific time.

        Parameters
        ----------
        data_key : str
            Data key.
        time : datetime.datetime
            Version time, by default None.
        show_all : bool
            show non-default columns, by default False

        Returns
        -------
        dict
            Version information.

        Raises
        ------
        RuntimeError
            If no version is defined at specified time.
        """
        version = self.get_data_version(data_key, time)

        return [c["name"] for c in version["columns"] if (c["default"] | show_all)]

    @property
    def data_keys(self):
        return list(self.spacecraft["data_keys"].keys())

    def resolve_data_key(self, data_key):
        """Replace data_key with actual key in case it is an alternative name.

        Parameters
        ----------
        data_key : str
            Data key (alternative).

        Returns
        -------
        str
            Data key.

        Raises
        ------
        ValueError
            If invalid time range is given.
        """
        # check if data_key exists, or check for general name
        if data_key not in self.spacecraft["data_keys"]:
            key_resolved = False

            for key in self.spacecraft["data_keys"]:
                alt_names = self.spacecraft["data_keys"][key].get("alt_keys", [])

                if data_key in alt_names:
                    data_key = key
                    key_resolved = True
                    break

            if not key_resolved:
                raise KeyError("data_key \"%s\" is not defined", data_key)

        return data_key


def read_task(args):
    """Wrapper for reading data files.

    Parameters
    ----------
    args : tuple
        Function arguments as tuple.

    Returns
    -------
    (list[float], np.ndarray)
        Evaluation datetimes as timestamps & processed data array.

    Raises
    ------
    KeyError
        If a data column is invalid.
    NotImplementedError
        If time format for text files is not supported.
    """
    (file_path, file_version, range_start, range_end, columns, frame, frame_cadence, kernels) = args

    column_dicts = []

    # resolve default columns
    if columns[0] == "~":
        default_columns = []

        for column in file_version["columns"]:
            if column.get("default", False):
                default_columns.append(column["name"])

        if len(columns) > 1:
            default_columns.extend(columns[1:])

        columns = default_columns

    # resolve alternative columns and populate column_dicts
    for i in range(len(columns)):
        valid_column = False

        for j in range(len(file_version["columns"])):
            column = file_version["columns"][j]

            if columns[i] != column["name"] and columns[i] in column.get("alt_names", []):
                columns[i] = column["name"]
                column_dicts.append(column)
                valid_column = True

                break
            elif columns[i] == column["name"]:
                column_dicts.append(column)
                valid_column = True

                break

        if not valid_column:
            raise KeyError("data column \"%s\" is invalid (%s)", columns[i], file_path)

    if "_cdf" in file_version["format"]:
        time, data = read_cdf_task(file_path, range_start, range_end, file_version, column_dicts,
                                   cdf_type=file_version["format"])
    elif file_version["format"] == "text":
        time, data = read_text_task(file_path, range_start, range_end, file_version, column_dicts)
    elif file_version["format"] == "fits_header":
        time, data = read_fits_header_task(file_path, range_start, range_end, file_version,
                                           column_dicts)
    else:
        raise NotImplementedError("format \"%s\" is not implemented", file_version["format"])

    # time mask
    time_mask = np.where((time > range_start.timestamp()) & (time < range_end.timestamp()))[0]

    if len(time_mask) == 1:
        raise NotImplementedError
    if len(time_mask) > 1:
        time = time[time_mask]

        # process data
        for i in range(len(data)):
            column = column_dicts[i]

            data_entry = data[i][time_mask]

            # filter values outside of range
            valid_range = column.get("valid_range", None)

            if valid_range:
                data_entry = np.where((data_entry > valid_range[0]) & (data_entry < valid_range[1]),
                                      data_entry, np.nan)

            sort_mask = np.argsort(time)
            time = time[sort_mask]
            data_entry = data_entry[sort_mask]

            if data_entry.ndim == 1:
                data_entry = data_entry.reshape((-1, 1))

            if frame and "frame" in column and frame != column.get("frame", None):
                heliosat.spice.spice_init()
                heliosat.spice.spice_reload(kernels)
                data_entry = transform_pos(time, data_entry, column["frame"], frame,
                                           frame_cadence)

            data[i] = data_entry

        return time, np.concatenate(data, axis=1)
    else:
        return None


def read_cdf_task(file_path, range_start, range_end, version_dict, column_dicts, cdf_type):
    """Worker function for reading cdf data files.

    Parameters
    ----------
    file_path : str
        File path.
    range_start : datetime.datetime
        Time range start datetime.
    range_end : datetime.datetime
        Time range end datetime.
    version_dict : dict
        Version information.
    column_dicts : list[dict]
        Data column information.
    cdf_type : str
        CDF file type (cdf, netcdf4).

    Returns
    -------
    (list[float], np.ndarray)
        Evaluation datetimes as timestamps & processed data array.
    """
    time_dict = version_dict["time"]
    logger = logging.getLogger(__name__)

    try:
        if cdf_type == "nasa_cdf":
            file = cdflib.CDF(file_path)
            epochs = file.varget(time_dict["key"])

            # fix for some cdf files that have 0 entries
            if np.sum(epochs == 0) > 0:
                null_filter = (epochs != 0)
                epochs = epochs[null_filter]
            else:
                null_filter = None

            time = cdflib.epochs.CDFepoch.unixtime(epochs, to_np=True)
            data = []

            for column in column_dicts:
                key = column["key"]

                if isinstance(key, str):
                    indices = column.get("indices", None)

                    if indices is not None:
                        indices = np.array(indices)
                        data.append(np.array(file.varget(key)[:, indices]))
                    else:
                        data.append(np.array(file.varget(key)))
                elif isinstance(key, list):
                    data.append(np.stack(arrays=[np.array(file.varget(k))
                                                 for k in key], axis=1))
                else:
                    raise NotImplementedError

            if null_filter is not None:
                for i in range(len(data)):
                    data[i] = data[i][null_filter]
        elif cdf_type == "net_cdf4":
            file = Dataset(file_path, "r")

            time = np.array([t / 1000 for t in file.variables[time_dict["key"]][...]])
            data = []

            for column in column_dicts:
                key = column["key"]

                if isinstance(key, str):
                    indices = column.get("indices", None)

                    if indices is not None:
                        indices = np.array(indices)
                        data.append(np.array(file[key][:, indices]))
                    else:
                        data.append(np.array(file[key][:]))
                elif isinstance(key, list):
                    data.append(np.stack(arrays=[np.array(file[k][:])
                                                 for k in key], axis=1))
                else:
                    raise NotImplementedError
        else:
            raise NotImplementedError("CDF type \"%s\" is not supported", cdf_type)

        return time, data
    except Exception as err:
        logger.error("failed to read file \"%s\" (%s)", file_path, err)
        raise Exception


def read_fits_header_task(file_path, range_start, range_end, version_dict, column_dicts):
    """Worker function for reading FITS files (headers).

    Parameters
    ----------
    file_path : str
        File path.
    range_start : datetime.datetime
        Time range start datetime.
    range_end : datetime.datetime
        Time range end datetime.
    version_dict : dict
        Version information.
    column_dicts : list[dict]
        Data column information.

    Returns
    -------
    (list[float], np.ndarray)
        Evaluation datetimes as timestamps & processed data array.
    """
    time_key = version_dict["time"]["key"]

    file = fits.open(file_path)

    time = np.array([string_to_datetime(file[i].header[time_key]).timestamp()
                     for i in range(1, len(file))])

    data = []

    for column in column_dicts:
        data.append(np.array([file[i].header[column["key"]] for i in range(1, len(file))]))

    return time, data


def read_text_task(file_path, range_start, range_end, version_dict, column_dicts):
    """Worker function for reading text files.

    Parameters
    ----------
    file_path : str
        File path.
    range_start : datetime.datetime
        Time range start datetime.
    range_end : datetime.datetime
        Time range end datetime.
    version_dict : dict
        Version information.
    column_dicts : list[dict]
        Data column information.

    Returns
    -------
    (list[float], np.ndarray)
        Evaluation datetimes as timestamps & processed data array.
    """
    logger = logging.getLogger(__name__)

    skip_rows = version_dict.get("skip_rows", 0)

    time_format = version_dict["time"]["format"]
    time_index = version_dict["time"]["index"]

    delimiter = version_dict.get("delimiter", None)

    def decode(string, format):
        if string.endswith("60.000"):
            string = "{0}59.000".format(string[:-6])

            return string_to_datetime(string, format).timestamp() + 1
        else:
            return string_to_datetime(string, format).timestamp()

    converters = {}

    skip_cols = version_dict.get("skip_columns", [])

    for col in skip_cols:
        converters[col] = lambda string: -1

    if isinstance(time_format, str):
        converters[0] = lambda string: decode(string, time_format)

        file = np.loadtxt(file_path, skiprows=skip_rows, encoding="latin1", delimiter=delimiter,
                          converters=converters)

        time = file[:, time_index]
    elif isinstance(time_format, int):
        file = np.loadtxt(file_path, skiprows=skip_rows, converters=converters)

        time = file[:, time_index] + time_format
    else:
        logger.exception("time_format \"%s\" is not implemented", type(time_format))
        raise NotImplementedError("time_format \"%s\" is not implemented", time_format)

    data = []

    for column in column_dicts:
        indices = column["indices"]

        if isinstance(indices, int):
            indices = column.get("indices", None)
            indices = np.array(indices)
            data.append(np.stack([np.array(file[:, index]) for index in indices]))

        elif isinstance(indices, list):
            data.append(np.stack([np.array(file[:, index])
                                  for index in indices], axis=1))
        else:
            raise NotImplementedError

    return time, data



# -*- coding: utf-8 -*-

"""util.py

General utility functions for the HelioSat package. For internal use only.
"""

import datetime
import logging
import os


from typing import Iterable, List, Optional, Union


def datetime_utc(*args: tuple) -> datetime.datetime:
    """Create tz-aware datetime object (UTC).
    """
    return datetime.datetime(*args).replace(tzinfo=datetime.timezone.utc)


def datetime_utc_timestamp(timestamp: float) -> datetime.datetime:
    """Create tz-aware datetime object (UTC) from timestamp.
    """
    return datetime.datetime.fromtimestamp(timestamp, datetime.timezone.utc)


def datetime_to_string(time_datetime: datetime.datetime) -> str:
    """Convert python datetime to string.

    Parameters
    ----------
    time_datetime : datetime.datetime
        Datetime object.

    Returns
    -------
    str
        Datetime string.
    """
    return time_datetime.strftime("%Y-%m-%dT%H:%M:%S.%f")


def get_heliosat_paths() -> dict:
    """Get heliosat file paths.

    Returns
    -------
    dict
        Dictionary containing heliosat paths.
    """
    root_path = os.getenv('HELIOSAT_DATAPATH', os.path.join(os.path.expanduser("~"), ".heliosat"))

    return {
        "cache": os.path.join(root_path, "cache"),
        "data": os.path.join(root_path, "data"),
        "kernels": os.path.join(root_path, "kernels"),
        "root": root_path
    }


def sanitize_datetimes(t: Union[datetime.datetime, Iterable[datetime.datetime]]) \
        -> Union[datetime.datetime, Iterable[datetime.datetime]]:
    """Sanitize tz-unaware datetime objects. Any tz-unaware datetime objects are set to UTC.

    Parameters
    ----------
    t : Union[datetime.datetime, Iterable[datetime.datetime]]
        Datetime objects to sanitize.

    Returns
    -------
    Union[datetime.datetime, Iterable[datetime.datetime]]
        Sainitized datetime objects.
    """
    if isinstance(t, datetime.datetime) and t.tzinfo is None:
        return t.replace(tzinfo=datetime.timezone.utc)
    elif isinstance(t, datetime.datetime):
        return t.astimezone(datetime.timezone.utc)
    elif hasattr(t, "__iter__"):
        _t = list(t)

        for i in range(len(_t)):
            if _t[i].tzinfo is None:
                _t[i] = _t[i].replace(tzinfo=datetime.timezone.utc)
            else:
                _t[i] = _t[i].astimezone(datetime.timezone.utc)

        return _t
    else:
        return t


def string_to_datetime(time_string: str, time_format: Optional[str] = None) -> datetime.datetime:
    """Convert string to python datetime object.

    Parameters
    ----------
    time_string : str
        Datetime string.
    time_format : str
        Datetime string format.

    Returns
    -------
    datetime.datetime
        Datetime object.

    Raises
    ------
    ValueError
        If string format is unkown.
    """
    formats = [
        "%Y-%m-%dT%H:%M:%S.%f",
        "%Y-%m-%dT%H:%M:%S",
        "%Y-%m-%dT%H:%MZ",
        "%Y-%m-%dT%H:%M",
        "%Y-%m-%d"
    ]

    if time_format:
        try:
            return datetime.datetime.strptime(
                time_string, time_format
                ).replace(tzinfo=datetime.timezone.utc)
        except ValueError:
            pass

    for f in formats:
        try:
            return datetime.datetime.strptime(time_string, f).replace(tzinfo=datetime.timezone.utc)
        except ValueError:
            pass

    raise ValueError("could not convert \"%s\", unkown format", time_string)



# -*- coding: utf-8 -*-

"""caching.py

Implements simple caching functionality. For internal use only.

Notes: All cached files are stored in a single folder located at either ~/.heliosat/cache or
$HELIOSAT_DATAPATH/cache if the environment variable is set. The required disk space can grow to be
very large if used extensively, but the folder can be deleted manually if required without the loss
of any functionality.
"""

import hashlib
import heliosat
import json
import logging
import os
import pickle


def cache_add_entry(key: str, obj: object):
    """Store data in cache using the specified key.

    Parameters
    ----------
    key : str
        Cache key.
    obj : object
        Data object.
    """
    logger = logging.getLogger(__name__)

    cache_path = heliosat._paths["cache"]

    if not os.path.exists(cache_path):
        logger.debug("cache path does not exist, creating \"%s\"", cache_path)
        os.makedirs(cache_path)

    with open(os.path.join(cache_path, "{}.cache".format(key)), "wb") as pickle_file:
        logger.debug("creating cache entry \"%s\"", key)
        pickle.dump(obj, pickle_file)


def cache_delete_entry(key: str):
    """Delete cache entry belonging to specified cache key.

    Parameters
    ----------
    key : str
        Cache key.

    Raises
    ------
    KeyError
        No entry for the given key is found.
    """
    logger = logging.getLogger(__name__)

    cache_path = heliosat._paths["cache"]

    if not cache_entry_exists(key):
        logger.exception("cache entry \"%s\" does not exist", key)
        raise KeyError("cache entry \"%s\" does not exist", key)

    logger.debug("deleting cache entry \"%s\"", key)
    os.remove(os.path.join(cache_path, "{0}.cache".format(key)))


def cache_entry_exists(key: str) -> bool:
    """Check if an entry for the given cache key exists.

    Parameters
    ----------
    key : str
        Cache key.

    Returns
    -------
    bool
        Flag if key exists.
    """
    cache_path = heliosat._paths["cache"]

    return os.path.exists(os.path.join(cache_path, "{}.cache".format(key)))


def cache_generate_key(identifiers: dict) -> str:
    """Generate SHA256 digest from the identifiers dictionary which acts as the cache key.

    Parameters
    ----------
    identifiers : dict
        Dictionary containing cache identifiers.

    Returns
    -------
    str
        Cache key (SHA256 digest).
    """
    hashobj = hashlib.sha256()
    hashobj.update(json.dumps(identifiers, sort_keys=True).encode("utf-8"))

    return hashobj.hexdigest()


def cache_get_entry(key: str) -> object:
    """Retrieve cache entry specified by cache key.

    Parameters
    ----------
    key : str
        Cache key.

    Returns
    -------
    object
        Data object.

    Raises
    ------
    KeyError
        No entry is found.
    """
    logger = logging.getLogger(__name__)

    cache_path = heliosat._paths["cache"]

    if not cache_entry_exists(key):
        logger.exception("cache key \"%s\" does not exist", key)
        raise KeyError("cache key \"%s\" does not exist", key)

    with open(os.path.join(cache_path, "{0}.cache".format(key)), "rb") as pickle_file:
        logger.debug("loading cache entry \"%s\"", key)
        cache_data = pickle.load(pickle_file)

    return cache_data



# -*- coding: utf-8 -*-

"""coordinates.py
Implements coordinate transformation functions requiring SPICE.
"""

import datetime
import heliosat
import multiprocessing
import numpy as np
import spiceypy

from concurrent.futures import ProcessPoolExecutor
from typing import Iterable, Optional, Union


def transform_lonlat(t: Union[datetime.datetime, Iterable[datetime.datetime]],
                     lonlat: np.ndarray, frame_from: str, frame_to: str) -> np.ndarray:
    """Transform longitude/latitude's from one reference frame to another.
    Inclination is not transformed, therefore the transformation only works in between reference
    frames that share the same z-axis.
    Parameters
    ----------
    t : Union[datetime.datetime, Iterable[datetime.datetime]]
        Evaluation datetimes.
    lonlat : np.ndarray
        Longitude/latitude's array.
    frame_from : str
        Source refernce frame.
    frame_to : str
        Target reference frame.
    Returns
    -------
    np.ndarray
        Transformed longitude/latitude's.
    Raises
    ------
    TypeError
        Variable t is not a valid datetime or Iterable of datetimes
    ValueError
        Source and target frame are equal.
    """
    if frame_from == frame_to:
        raise ValueError("source and target frame are equal")

    lons_rad = 2 * np.pi * lonlat[:, 0] / 360
    lats_rad = 2 * np.pi * lonlat[:, 1] / 360

    vecs = np.array([
        [np.cos(lon) * np.cos(lat), np.sin(lon) * np.cos(lat), np.sin(lat)]
        for (lon, lat) in list(zip(lons_rad, lats_rad))
    ])

    if isinstance(t, datetime.datetime):
        for i in range(0, len(vecs)):
            vecs[i] = spiceypy.mxv(spiceypy.pxform(frame_from, frame_to,
                                   spiceypy.datetime2et(t)),
                                   vecs[i])
    elif hasattr(t, "__iter__"):
        for i in range(0, len(t)):
            vecs[i] = spiceypy.mxv(spiceypy.pxform(frame_from, frame_to,
                                   spiceypy.datetime2et(t[i])),
                                   vecs[i])
    else:
        raise TypeError("Variable t is not a valid datetime or Iterable of datetimes")

    return np.array([
        [360 * np.arccos(v[0]) / 2 / np.pi, 360 * np.arcsin(v[2]) / 2 / np.pi]
        for v in vecs
    ])


def transform_pos(t: Union[datetime.datetime, Iterable[datetime.datetime]],
                  data: np.ndarray, frame_from: str, frame_to: str,
                  frame_cadence: Optional[float] = None) -> np.ndarray:
    """Transform 3D coordinates from one reference frame to another.
    Parameters
    ----------
    t : Union[datetime.datetime, Iterable[datetime.datetime]]
        Evaluation datetimes.
    data : np.ndarray
        3D vector array.
    frame_from : str
        Source refernce frame.
    frame_to : str
        Target reference frame.
    frame_cadence: float, optional
        Evaluate frame transformation matrix every "frame_cadence" seconds instead of at very
        time point (significant speed up), by default None.
    Returns
    -------
    np.ndarray
        Transformed vector array in target reference frame.
    Raises
    ------
    ValueError
        Data array has the wrong dimensions (must be 2d or 3d)
        Source and target frame are equal.
    """
    if frame_from == frame_to:
        return data

    # convert timestamps to python datetimes if required
    if not isinstance(t[0], datetime.datetime):
        t = [datetime.datetime.fromtimestamp(_t, datetime.timezone.utc) for _t in t]

    if frame_cadence:
        frames = int((t[-1] - t[0]).total_seconds() // frame_cadence)
        frame_indices = [np.floor(_) for _ in np.linspace(0, frames, len(t), endpoint=False)]
        time_indices = np.linspace(0, len(t), frames, endpoint=False)

        frames = [spiceypy.pxform(frame_from, frame_to, spiceypy.datetime2et(t[int(i)]))
                  for i in time_indices]
    else:
        frames = None
        frame_indices = None

    if data.ndim == 2:
        if frame_cadence:
            for i in range(0, len(t)):
                data[i] = spiceypy.mxv(frames[int(frame_indices[i])], data[i])
        else:
            for i in range(0, len(t)):
                data[i] = spiceypy.mxv(spiceypy.pxform(frame_from, frame_to,
                                       spiceypy.datetime2et(t[i])),
                                       data[i])

        return data
    elif data.ndim == 3:
        max_workers = min(multiprocessing.cpu_count() * 2, data.shape[1])
        kernels = heliosat._spice["kernels_loaded"]

        with ProcessPoolExecutor(max_workers=max_workers) as executor:
            args = [(t, data[:, i], frame_from, frame_to, frames, frame_indices, kernels)
                    for i in range(data.shape[1])]

            futures = executor.map(_worker_transform_pos, args)

        result = np.array([_ for _ in futures])

        return np.swapaxes(result, 0, 1)

    else:
        raise ValueError("data array can only be 2 or 3 dimensional")


def _worker_transform_pos(args: tuple) -> np.ndarray:
    """Worker function for transforming 3D coordinates.
    Parameters
    ----------
    args: (np.ndarray, np.ndarray, str, str, np.ndarray, np.ndarray, list)
        Function arguments as tuple.
    Returns
    -------
    np.ndarray
        Transformed 3D coordinates.
    """
    (t, data, frame_from, frame_to, frames, frame_indices, kernels) = args

    heliosat.spice.spice_init()
    heliosat.spice.spice_reload(kernels)

    if frames:
        for i in range(0, len(t)):
            data[i] = spiceypy.mxv(frames[int(frame_indices[i])], data[i])
    else:
        for i in range(0, len(t)):
            data[i] = spiceypy.mxv(spiceypy.pxform(frame_from, frame_to,
                                   spiceypy.datetime2et(t[i])),
                                   data[i])
    return data



# -*- coding: utf-8 -*-

"""smoothing.py

Implements smoothing functionality.
"""

import datetime
import logging
import numba
import numpy as np

from typing import Iterable


def smooth_data(t: Iterable[datetime.datetime],
                time_raw: np.ndarray, data_raw: np.ndarray, **kwargs) -> (np.ndarray, np.ndarray):
    """Smooth raw data and evaluate at timesteps t.

    Parameters
    ----------
    t : Iterable[datetime.datetime]
        Evaluation datetimes.
    time_raw : np.ndarray
        Raw time array.
    data_raw : np.ndarray
        Raw data array.

    Other Parameters
    ----------------
    smoothing: str
        Smoothing method, by default "kernel".
    smoothing_scale: float
        Smoothing scale in seconds, by default 300.

    Returns
    -------
    (np.ndarray, np.ndarray)
            Evaluation datetimes as timestamps & smoothed data array.

    Raises
    ------
    NotImplementedError
        If smoothing method is not implemented.
    """
    logger = logging.getLogger(__name__)

    time_smooth = np.array([_t.timestamp() for _t in t])
    data_smooth = np.zeros((len(t), data_raw.shape[1]))

    smoothing = kwargs.get("smoothing", "kernel")
    smoothing_scale = kwargs.get("smoothing_scale", 300)

    if smoothing in ["average", "moving_average", "mean"]:
        _smoothing_mean(time_smooth, time_raw, data_raw, data_smooth, smoothing_scale)
    elif smoothing in ["kernel", "kernel_gaussian", "gaussian"]:
        _smoothing_gaussian_kernel(time_smooth, time_raw, data_raw, data_smooth, smoothing_scale)
    elif smoothing in ["linear", "linear_interpolation"]:
        data_smooth = np.array([np.interp(time_smooth, time_raw, data_raw[:, i])
                                for i in range(data_raw.shape[1])])
        return time_smooth, data_smooth.T
    elif smoothing in ["closest"]:
        return _smoothing_closest(time_smooth, time_raw, data_raw, data_smooth)
    else:
        logger.exception("smoothing method \"%s\" is not implemented", kwargs.get("smoothing"))
        raise NotImplementedError("smoothing method \"%s\" is not implemented",
                                  kwargs.get("smoothing"))

    return time_smooth, data_smooth


@numba.njit(parallel=False)
def _smoothing_closest(t, time_raw, data_raw, data_smooth):
    """Smooth data using closest.

    Parameters
    ----------
    t : list[float]
        Evaluation times as timestamps.
    time_raw : np.ndarray
        Raw time array.
    data_raw : np.ndarray
        Raw data array.
    data_smooth : np.ndarray
        Smoothed data array.
    smoothing_scale : float
        Smoothing scale in seconds.
    """
    t_actual = np.zeros_like(t)

    for i in numba.prange(len(t)):
        index = np.argmin(np.abs(t[i] - time_raw))

        t_actual[i] = time_raw[index]
        data_smooth[i] = data_raw[index]

    return t_actual, data_smooth


@numba.njit(parallel=True)
def _smoothing_mean(t, time_raw, data_raw, data_smooth, smoothing_scale):
    """Smooth data using moving average.

    Parameters
    ----------
    t : list[float]
        Evaluation times as timestamps.
    time_raw : np.ndarray
        Raw time array.
    data_raw : np.ndarray
        Raw data array.
    data_smooth : np.ndarray
        Smoothed data array.
    smoothing_scale : float
        Smoothing scale in seconds.
    """
    for i in numba.prange(len(t)):
        total = 0
        dims = data_raw.shape[1]
        vector = np.zeros((dims,))

        for j in range(0, len(data_raw)):
            if np.abs(time_raw[j] - t[i]) < smoothing_scale:
                total += 1

                for k in range(0, len(vector)):
                    vector[k] += data_raw[j, k]

        for k in range(0, len(vector)):
            if total == 0:
                data_smooth[i, k] = np.nan
            else:
                data_smooth[i, k] = vector[k] / total


@numba.njit(parallel=True)
def _smoothing_gaussian_kernel(t, time_raw, data_raw, data_smooth, smoothing_scale):
    """Smooth data using a gaussian kernel.

    Parameters
    ----------
    t : list[float]
        Evaluation times as timestamps.
    time_raw : np.ndarray
        Raw time array.
    data_raw : np.ndarray
        Raw data array.
    data_smooth : np.ndarray
        Smoothed data array.
    smoothing_scale : float
        Smoothing scale in seconds.
    """
    for i in numba.prange(len(t)):
        total = 0
        dims = data_raw.shape[1]
        vector = np.zeros((dims,))

        for j in range(0, len(data_raw)):
            if np.abs(time_raw[j] - t[i]) < 3 * smoothing_scale and not np.isnan(data_raw[j, 0]):
                kernel = np.exp(-(time_raw[j] - t[i]) ** 2 / 2 / smoothing_scale ** 2)

                total += kernel

                for k in range(0, len(vector)):
                    vector[k] += kernel * data_raw[j, k]

        for k in range(0, len(vector)):
            if total == 0:
                data_smooth[i, k] = np.nan
            else:
                data_smooth[i, k] = vector[k] / total



# -*- coding: utf-8 -*-

"""satellites.py

Implements spacecraft classes.
"""

import datetime
import logging
import heliosat
import inspect
import spiceypy

from heliosat.coordinates import transform_pos
from heliosat.spacecraft import Spacecraft
from typing import Iterable, List, Optional, Union


class BEPI(Spacecraft):
    def __init__(self, **kwargs):
        super(BEPI, self).__init__("bepi_colombo", body_name="BEPICOLOMBO MPO", **kwargs)


class DSCOVR(Spacecraft):
    def __init__(self, **kwargs):
        super(DSCOVR, self).__init__("dscovr", body_name="EARTH", **kwargs)


class MES(Spacecraft):
    def __init__(self, **kwargs):
        super(MES, self).__init__("messenger", body_name="MESSENGER", **kwargs)


class PSP(Spacecraft):
    def __init__(self, **kwargs):
        from spiceypy.utils.support_types import SpiceyError

        super(PSP, self).__init__("psp", body_name="SPP", **kwargs)

        try:
            spiceypy.bodn2c("SPP_SPACECRAFT")
        except SpiceyError:
            # kernel fix that is required due to PSP having different names
            spiceypy.boddef("SPP_SPACECRAFT", spiceypy.bodn2c("SPP"))


class SOLO(Spacecraft):
    def __init__(self, **kwargs):
        super(SOLO, self).__init__("solar_orbiter", body_name="SOLAR ORBITER", **kwargs)


class STA(Spacecraft):
    def __init__(self, **kwargs):
        super(STA, self).__init__("stereo_ahead", body_name="STEREO AHEAD", **kwargs)


class STB(Spacecraft):
    def __init__(self, **kwargs):
        super(STB, self).__init__("stereo_behind", body_name="STEREO BEHIND", **kwargs)


class VEX(Spacecraft):
    def __init__(self, **kwargs):
        super(VEX, self).__init__("venus_express", body_name="VENUS EXPRESS", **kwargs)


class WIND(Spacecraft):
    def __init__(self, **kwargs):
        super(WIND, self).__init__("wind", body_name="EARTH", **kwargs)

    def trajectory(self, t: Union[datetime.datetime, Iterable[datetime.datetime]],
                   frame: str = "J2000", observer: str = "SUN", units: str = "AU"):
        """Evaluate body trajectory at given datetimes.

        Parameters
        ----------
        t : Union[datetime.datetime, Iterable[datetime.datetime]]
            Evaluate datetime(s).
        frame : str, optional
            Trajectory reference frame, by default "J2000".
        observer : str, optional
            Observer body name, by default "SUN".
        units : str, optional
            Output units, by default "AU".

        Returns
        -------
        np.ndarray
            Body trajectory.

        Raises
        ------
        NotImplementedError
            If units are invalid.
        """
        logger = logging.getLogger(__name__)

        if heliosat._spice is None:
            logger.info("running spice_init")
            spice_init(False)

        if isinstance(t, datetime.datetime):
            traj_t, traj_p = self.get_data([t, t + datetime.timedelta(hours=1)], "wind_trajectory", cache=True)
            traj_p = traj_p[0]
            traj_p[0] *= -1
            traj_p[1] *= -1

            traj_p += heliosat.Earth().trajectory(t, units="km", frame="HEE")

            trajectory = spiceypy.mxv(spiceypy.pxform("HEE", frame, spiceypy.datetime2et(t)), traj_p)
        elif len(t) == 2:
            traj_t, traj_p = self.get_data([t[0], t[1], t[1] + datetime.timedelta(hours=1)], "wind_trajectory", cache=True)
            traj_p = traj_p[0:2]

            traj_p[:, 0] *= -1
            traj_p[:, 1] *= -1

            traj_p += heliosat.Earth().trajectory(t, units="km", frame="HEE")
            
            trajectory = transform_pos(t, traj_p, "HEE", frame)
        else:
            traj_t, traj_p = self.get_data(t, "wind_trajectory", cache=True)
            traj_p[:, 0] *= -1
            traj_p[:, 1] *= -1

            traj_p += heliosat.Earth().trajectory(t, units="km", frame="HEE")
            
            trajectory = transform_pos(t, traj_p, "HEE", frame)

        if units == "AU":
            trajectory *= 6.68459e-9
        elif units == "m":
            trajectory *= 1e3
        elif units == "km":
            pass
        else:
            logger.exception("unit \"%s\" is not supported", units)
            raise NotImplementedError("unit \"%s\" is not supported", units)

        return trajectory


def select_satellite(satellite: str) -> Spacecraft:
    if hasattr(heliosat, satellite):
        sat = getattr(heliosat, satellite)

        if Spacecraft in inspect.getmro(sat):
            return sat

    raise NotImplementedError("unkown satellite \"%s\"", satellite)



# -*- coding: utf-8 -*-

from setuptools import setup


setup(
    name="HelioSat",
    packages=[
        "heliosat"
    ],
    package_data={"heliosat": ["json/*.json"]},
    version="0.4.10",
    author="Andreas J. Weiss",
    author_email="andreas.weiss@oeaw.ac.at",
    keywords=["astrophysics", "solar physics", "space weather"],
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/ajefweiss/HelioSat",
    install_requires=[
        "astropy",
        "bs4",
        "cdflib>=0.3.19",
        "netcdf4",
        "numba",
        "numpy",
        "requests",
        "spiceypy"
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics",
    ],
)
